/**
 * Copyright (c) 2017-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 * 
 */

'use strict';
Object.defineProperty(exports, '__esModule', {value: true});

var _net = require('net');
var _net2 = _interopRequireDefault(_net);
var _rsocketFlowable = require('rsocket-flowable');
var _Deferred = require('fbjs/lib/Deferred');
var _Deferred2 = _interopRequireDefault(_Deferred);
var _rsocketCore = require('rsocket-core');
function _interopRequireDefault(obj) {
  return obj && obj.__esModule ? obj : {default: obj};
}

/**
                                                                                                                                          * A TCP transport client for use in node environments.
                                                                                                                                          */
class RSocketTcpClient {
  constructor(options, Encoders) {
    this._encoders = Encoders;
    this._options = options;
  }

  connect() {
    return new _rsocketFlowable.Single(subscriber => {
      const socket = _net2.default.connect(this._options);
      const onError = error => {
        socket.removeAllListeners();
        subscriber.onError(error);
      };
      const onComplete = () => {
        socket.removeAllListeners();
        subscriber.onComplete(
          new TcpDuplexConnection(this._options, this._encoders, socket)
        );
      };
      const remove = () => {
        socket.removeAllListeners();
        socket.end();
      };

      subscriber.onSubscribe(remove);
      socket.once('error', onError);
      socket.once('connect', onComplete);
    });
  }
}
exports.default = RSocketTcpClient;

/**
                                         * @private
                                         */
class TcpDuplexConnection {
  constructor(options, encoders, socket) {
    this.close = () => {
      if (!this._active) {
        return;
      }
      this._active = false;
      this._close.resolve();
      this._receivers.forEach(subscriber => subscriber.onComplete());
      this._receivers.clear();
      this._senders.forEach(subscription => subscription.cancel());
      this._senders.clear();
      this._socket.removeAllListeners();
      this._socket.end();
    };
    this._handleData = chunk => {
      try {
        const frames = this._readFrames(chunk);
        frames.forEach(frame => {
          this._receivers.forEach(subscriber => subscriber.onNext(frame));
        });
      } catch (error) {
        this._handleError(error);
        this.close();
      }
    };
    this._handleEnd = () => {
      this._handleError(
        new Error('RSocketTcpClient: Socket closed unexpectedly.')
      );
    };
    this._handleError = error => {
      this._receivers.forEach(subscriber => subscriber.onError(error));
      this.close();
    };
    this._active = true;
    this._buffer = (0, _rsocketCore.createBuffer)(0);
    this._close = new _Deferred2.default();
    this._encoders = encoders;
    this._options = options;
    this._receivers = new Set();
    this._senders = new Set();
    this._socket = socket;
    this._socket.on('close', this.close);
    this._socket.on('data', this._handleData);
    this._socket.on('end', this._handleEnd);
    this._socket.on('error', this._handleError);
  }
  sendOne(frame) {
    this._writeFrame(frame);
  }
  send(frames) {
    let subscription;
    frames.subscribe({
      onComplete: () => {
        subscription && this._senders.delete(subscription);
      },
      onError: error => this._handleError(error),
      onNext: frame => this._writeFrame(frame),
      onSubscribe: _subscription => {
        subscription = _subscription;
        subscription.request(Number.MAX_SAFE_INTEGER);
        this._senders.add(subscription);
      },
    });
  }
  receive() {
    return new _rsocketFlowable.Flowable(subject => {
      subject.onSubscribe({
        cancel: () => {
          this._receivers.delete(subject);
        },
        request: () => {
          this._receivers.add(subject);
        },
      });
    });
  }
  onClose() {
    return this._close.getPromise();
  }

  _readFrames(chunk) {
    // Combine results from any partial frames received with new data and
    // extract any frames plus remaining bytes.
    const buffer = Buffer.concat([this._buffer, chunk]);
    const [frames, remaining] = (0, _rsocketCore.deserializeFrames)(
      buffer,
      this._encoders
    );
    this._buffer = remaining;
    return frames;
  }

  _writeFrame(frame) {
    try {
      const buffer = (0, _rsocketCore.serializeFrameWithLength)(
        frame,
        this._encoders
      );
      this._socket.write(buffer);
    } catch (error) {
      this._handleError(error);
    }
  }
}
